package dcos.metronome
package repository.impl.kv.marshaller

import java.util.concurrent.TimeUnit

import dcos.metronome.Protos
import dcos.metronome.model._
import dcos.metronome.repository.impl.kv.EntityMarshaller
import dcos.metronome.scheduler.TaskState
import mesosphere.marathon.core.task.Task
import org.joda.time.{ DateTime, DateTimeZone }
import org.slf4j.LoggerFactory

import scala.collection.JavaConverters._
import scala.concurrent.duration.Duration

object JobRunMarshaller extends EntityMarshaller[JobRun] {
  val log = LoggerFactory.getLogger(JobRunMarshaller.getClass)

  override def toBytes(jobRun: JobRun): IndexedSeq[Byte] = {
    import JobRunConversions.JobRunToProto

    jobRun.toProto.toByteArray
  }

  override def fromBytes(bytes: IndexedSeq[Byte]): Option[JobRun] =
    safeConversion { fromProto(Protos.JobRun.parseFrom(bytes.toArray)) }

  private def fromProto(proto: Protos.JobRun): JobRun = {
    import JobRunConversions.ProtoToJobRun

    proto.toModel
  }
}

object JobRunConversions {
  implicit class JobRunIdToProto(val jobRunId: JobRunId) extends AnyVal {
    def toProto: Protos.JobRun.Id = {
      Protos.JobRun.Id.newBuilder()
        .setJobId(jobRunId.jobId.toString)
        .setRunId(jobRunId.value)
        .build()
    }
  }

  implicit class JobRunTaskToProto(val task: JobRunTask) extends AnyVal {
    def toProto: Protos.JobRun.JobRunTask = {
      val builder = Protos.JobRun.JobRunTask.newBuilder()
        .setId(task.id.idString)
        .setStartedAt(task.startedAt.getMillis)
        .setStatus(task.status.toProto)
      task.completedAt.foreach(t => builder.setCompletedAt(t.getMillis))
      builder.build()
    }
  }

  implicit class ProtoToJobRunTask(val proto: Protos.JobRun.JobRunTask) extends AnyVal {
    def toModel: JobRunTask = JobRunTask(
      Task.Id(proto.getId),
      new DateTime(proto.getStartedAt, DateTimeZone.UTC),
      if (proto.hasCompletedAt) Some(new DateTime(proto.getCompletedAt, DateTimeZone.UTC)) else None,
      proto.getStatus.toModel)
  }

  implicit class ProtoToJobRunId(val proto: Protos.JobRun.Id) extends AnyVal {
    def toModel: JobRunId = JobRunId(JobId(proto.getJobId), proto.getRunId)
  }

  implicit class JobRunToProto(val jobRun: JobRun) extends AnyVal {
    def toProto: Protos.JobRun = {
      import JobSpecConversions.JobSpecToProto

      val builder = Protos.JobRun.newBuilder()
        .setId(jobRun.id.toProto)
        .setJobSpec(jobRun.jobSpec.toProto)
        .setStatus(jobRun.status.toProto)
        .setCreatedAt(jobRun.createdAt.getMillis)

      jobRun.completedAt.foreach(date => builder.setFinishedAt(date.getMillis))
      jobRun.tasks.values.foreach(task => builder.addTasks(task.toProto))
      jobRun.startingDeadline.foreach(d => builder.setStartingDeadlineSeconds(d.toSeconds))
      builder.build()
    }
  }

  implicit class ProtoToJobRun(val proto: Protos.JobRun) extends AnyVal {
    def toModel: JobRun = {
      import JobSpecConversions.ProtoToJobSpec

      JobRun(
        id = proto.getId.toModel,
        jobSpec = proto.getJobSpec.toModel,
        status = proto.getStatus.toModel,
        createdAt = new DateTime(proto.getCreatedAt, DateTimeZone.UTC),
        completedAt = if (proto.hasFinishedAt) Some(new DateTime(proto.getFinishedAt, DateTimeZone.UTC)) else None,
        startingDeadline = if (proto.hasStartingDeadlineSeconds)
          Some(Duration(proto.getStartingDeadlineSeconds, TimeUnit.SECONDS))
        else
          None,
        tasks = proto.getTasksList.asScala.map(_.toModel).map(task => task.id -> task).toMap)
    }
  }

  implicit class JobRunStatusToProto(val status: JobRunStatus) extends AnyVal {
    def toProto: Protos.JobRun.Status = Protos.JobRun.Status.valueOf(JobRunStatus.name(status))
  }

  implicit class ProtoToJobRunStatus(val proto: Protos.JobRun.Status) extends AnyVal {
    def toModel: JobRunStatus = JobRunStatus.names(proto.toString)
  }

  implicit class TaskStateToProto(val taskState: TaskState) extends AnyVal {
    def toProto: Protos.JobRun.JobRunTask.Status = Protos.JobRun.JobRunTask.Status.valueOf(TaskState.name(taskState))
  }

  implicit class ProtoToTaskState(val proto: Protos.JobRun.JobRunTask.Status) extends AnyVal {
    def toModel: TaskState = TaskState.names(proto.toString)
  }
}
